/*
 * This file is part of the FORS Library
 * Copyright (C) 2002-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */


#ifndef FORS_SATURATION_MOS_H
#define FORS_SATURATION_MOS_H

#include <cpl.h>
#include <vector>
#include "mosca_image.h"
#include "fors_calibrated_slits.h"

/* Saturation computation */
void fors_saturation_reject_sat_slits(std::vector<mosca::image>& raw_flats,
                                 const fors::calibrated_slits& slits,
                                 cpl_mask ** non_linear_flat_masks,
                                 cpl_mask ** saturated_flat_masks,
                                 double max_nonlinear_ratio,
                                 std::vector<std::vector<double> >& slit_nonlinear_ratio,
                                 std::vector<std::vector<int> >& slit_nonlinear_count);

#endif
